#include "database/database.h"
#include "utility/progress.h"

#include "expat/expat.h"

#include <qfile.h>
#include <qdom.h>
#include <qapplication.h>

#include <stdexcept>
#include <cstring>




// tSAXDatabaseHandler --------------------------------------------------------
class tSAXDatabaseHandler
{
    enum tMode {
      NONE,
      DIRECTORIES,
      HISTORY,
      SONGS,
      SONG_SET,
      SONG_SET_NODE,
      SONG_SET_RENDERING,
      SONG_SET_POSITIVE_SET,
      SONG_SET_NEGATIVE_SET
    } CurrentMode;

    tDatabase &Database;
    tProgress	*Progress;
    XML_Parser		Parser;

    typedef vector<tSongSetNode *> tSongSetNodeStack;
    tSongSetNodeStack	SongSetNodeStack;

    tUniqueIdList	UniqueIdList;

  public:
    tSAXDatabaseHandler( tDatabase &database, tProgress *progress,
	XML_Parser parser );

    void startElement( const XML_Char *name, const XML_Char **attributes );
    void endElement( const XML_Char *name);
};




tSAXDatabaseHandler::tSAXDatabaseHandler( tDatabase &database, tProgress *progress,
    XML_Parser parser )
  : CurrentMode( NONE ), Database( database ), Progress( progress ), Parser( parser )
{
}





void tSAXDatabaseHandler::startElement( const XML_Char *name, const XML_Char **attributes )
{
  if ( Progress ) 
  {
    int current_progress = XML_GetCurrentByteIndex( Parser );
    if ( current_progress - Progress->progress() > 100000 )
      Progress->setProgress( current_progress );
  }

  // items --------------------------------------------------------------------
  if ( strcmp( name, "song" ) == 0 )
  {
    switch ( CurrentMode )
    {
      case SONGS:
	Database.SongCollection.insert( deserializeSong( attributes ), true );
	return;
      case SONG_SET_RENDERING:
      case SONG_SET_NEGATIVE_SET:
      case SONG_SET_POSITIVE_SET:
	UniqueIdList.push_back( lookupAttribute( "unique_id", attributes ).toUInt() );
	return;
      default:
	;
    }
  }
  if ( strcmp( name, "directory" ) == 0 && CurrentMode == DIRECTORIES )
  {
    Database.DirectoryList.push_back( lookupAttribute( "name", attributes ) );
    return;
  }
  if ( strcmp( name, "playedsong" ) == 0 && CurrentMode == HISTORY )
  {
    Database.History.played(
	lookupAttribute( "uniqueid", attributes ).toUInt(),
	lookupAttribute( "time", attributes ).toUInt(),
	lookupAttribute( "playduration", attributes ).toDouble()
	);
    return;
  }

  // mode changes -------------------------------------------------------------
  if ( strcmp( name, "madmandb" ) == 0 && CurrentMode == NONE )
    return;
  if ( strcmp( name, "songs" ) == 0 && CurrentMode == NONE )
  {
    CurrentMode = SONGS;
    return;
  }
  if ( strcmp( name, "directories" ) == 0 && CurrentMode == NONE )
  {
    CurrentMode = DIRECTORIES;
    return;
  }
  if ( strcmp( name, "history" ) == 0 && CurrentMode == NONE )
  {
    CurrentMode = HISTORY;
    return;
  }
  if ( strcmp( name, "song_set_node" ) == 0 && ( CurrentMode == NONE || CurrentMode == SONG_SET_NODE ) )
  {
    CurrentMode = SONG_SET_NODE;
    tSongSetNode *my_node = new tSongSetNode( NULL );
    my_node->setName( lookupAttribute( "name", attributes ) );

    if ( SongSetNodeStack.size() == 0 )
      Database.SongSetTree = my_node;
    else
      SongSetNodeStack.back()->insertChild( 
	  my_node, SongSetNodeStack.back()->end() );

    SongSetNodeStack.push_back( my_node );
    return;
  }
  if ( strcmp( name, "song_set" ) == 0 && ( CurrentMode == SONG_SET_NODE ) )
  {
    CurrentMode = SONG_SET;
    tSongSet *my_set = new tSongSet;
    my_set->setSongCollection( &Database.SongCollection );
    SongSetNodeStack.back()->setData( my_set );
    my_set->setCriterion( lookupAttribute( "criterion", attributes ) );
    return;
  }
  if ( strcmp( name, "rendering" ) == 0 && ( CurrentMode == SONG_SET ) )
  {
    CurrentMode = SONG_SET_RENDERING;
    UniqueIdList.clear();
    return;
  }
  if ( strcmp( name, "positiveset" ) == 0 && ( CurrentMode == SONG_SET ) )
  {
    CurrentMode = SONG_SET_POSITIVE_SET;
    UniqueIdList.clear();
    return;
  }
  if ( strcmp( name, "negativeset" ) == 0 && ( CurrentMode == SONG_SET ) )
  {
    CurrentMode = SONG_SET_NEGATIVE_SET;
    UniqueIdList.clear();
    return;
  }

  // error condition ----------------------------------------------------------
  cerr << "error: " << QString::fromUtf8( name, strlen( name ) ) << endl;
  runtimeError( qApp->translate( "ErrorMessages", "Invalid database structure"  ) );
}




void tSAXDatabaseHandler::endElement( const XML_Char *name)
{
  if ( strcmp( name, "madmandb" ) == 0 && CurrentMode == NONE )
    return;
  if ( strcmp( name, "songs" ) == 0 && CurrentMode == SONGS )
  {
    CurrentMode = NONE;
    return;
  }
  if ( strcmp( name, "directories" ) == 0 && CurrentMode == DIRECTORIES )
  {
    CurrentMode = NONE;
    return;
  }
  if ( strcmp( name, "history" ) == 0 && CurrentMode == HISTORY )
  {
    CurrentMode = NONE;
    return;
  }
  if ( strcmp( name, "song_set_node" ) == 0 && CurrentMode == SONG_SET_NODE )
  {
    SongSetNodeStack.pop_back();
    if ( SongSetNodeStack.size() == 0 )
      CurrentMode = NONE;
    else
      CurrentMode = SONG_SET_NODE;

    return;
  }
  if ( strcmp( name, "song_set" ) == 0 && ( CurrentMode == SONG_SET ) )
  {
    CurrentMode = SONG_SET_NODE;
    return;
  }
  if ( strcmp( name, "rendering" ) == 0 && ( CurrentMode == SONG_SET_RENDERING ) )
  {
    CurrentMode = SONG_SET;
    SongSetNodeStack.back()->data()->setRendering( UniqueIdList );
    return;
  }
  if ( strcmp( name, "positiveset" ) == 0 && ( CurrentMode == SONG_SET_POSITIVE_SET ) )
  {
    CurrentMode = SONG_SET;
    SongSetNodeStack.back()->data()->setPositiveSet( UniqueIdList );
    return;
  }
  if ( strcmp( name, "negativeset" ) == 0 && ( CurrentMode == SONG_SET_NEGATIVE_SET ) )
  {
    CurrentMode = SONG_SET;
    SongSetNodeStack.back()->data()->setNegativeSet( UniqueIdList );
    return;
  }
}




// expat parsing code ---------------------------------------------------------
void startExpatElement(void *data, const XML_Char *el, const XML_Char **attr)
{
  reinterpret_cast<tSAXDatabaseHandler *>( data )->startElement( el, attr );
}




void endExpatElement( void *data, const XML_Char *el ) 
{
  reinterpret_cast<tSAXDatabaseHandler *>( data )->endElement( el );
} 




// tDatabase ------------------------------------------------------------------
tDatabase::tDatabase()
: SongSetTree( NULL )
{
}




tDatabase::~tDatabase()
{
  if ( SongSetTree )
    delete SongSetTree;
}




void tDatabase::clear()
{
  if ( SongSetTree )
  {
    delete SongSetTree;
    SongSetTree = NULL;
  }

  SongCollection.clear();
  DirectoryList.clear();
  History.clear();
}




void tDatabase::startNew()
{
  clear();
  SongSetTree = new tSongSetNode( new tSongSet );
  SongSetTree->data()->setSongCollection( &SongCollection );
  SongSetTree->setName( qApp->translate( "SongSetName", "Root Playlist" ) );

  tSongSetNode *node = new tSongSetNode( new tSongSet );
  node->data()->setSongCollection( &SongCollection );
  node->data()->setCriterion( "~rating(>=4)" );
  node->data()->reevaluateCriterion();
  node->setName( qApp->translate( "SongSetName", "Absolute favorites" ) );
  SongSetTree->addChild( node );
  
  node = new tSongSetNode( new tSongSet );
  node->data()->setSongCollection( &SongCollection );
  node->data()->setCriterion( "~album(love)|~title(love)" );
  node->data()->reevaluateCriterion();
  node->setName( qApp->translate( "SongSetName", "Love songs" ) );
  SongSetTree->addChild( node );
  
  node = new tSongSetNode( new tSongSet );
  node->data()->setSongCollection( &SongCollection );
  node->data()->setCriterion( "~genre(rock)" );
  node->data()->reevaluateCriterion();
  node->setName( qApp->translate( "SongSetName", "Rock songs" ) );
  SongSetTree->addChild( node );

  node = new tSongSetNode( new tSongSet );
  node->data()->setSongCollection( &SongCollection );
  node->data()->setCriterion( "~playcount(=0)" );
  node->data()->reevaluateCriterion();
  node->setName( qApp->translate( "SongSetName", "Never been listened" ) );
  SongSetTree->addChild( node );

  node = new tSongSetNode( new tSongSet );
  node->data()->setSongCollection( &SongCollection );
  node->data()->setCriterion( "~existedfordays(<30)" );
  node->data()->reevaluateCriterion();
  node->setName( qApp->translate( "SongSetName", "Recent additions" ) );
  SongSetTree->addChild( node );

  node = new tSongSetNode( new tSongSet );
  node->data()->setSongCollection( &SongCollection );
  node->data()->setCriterion( "~fullplayratio(>=0.5)" );
  node->data()->reevaluateCriterion();
  node->setName( qApp->translate( "SongSetName", "Often listened to the end" ) );
  SongSetTree->addChild( node );

  node = new tSongSetNode( new tSongSet );
  node->data()->setSongCollection( &SongCollection );
  node->data()->setCriterion( "~fullplayratio(<0.5)" );
  node->data()->reevaluateCriterion();
  node->setName( qApp->translate( "SongSetName", "Rarely listened to the end" ) );
  SongSetTree->addChild( node );
}




void tDatabase::load( const QString &filename, tProgress *progress )
{
  clear();
  QFile dbfile( filename );

  if ( dbfile.open( IO_ReadOnly ) )
  {
    if ( progress )
    {
      progress->setTotalSteps( dbfile.size() );
      progress->setProgress( 0 );
    }

    tSongCollectionScopedBulkChange bulkchange( SongCollection );

    XML_Parser parser = XML_ParserCreate( NULL );
    if ( parser == NULL )
      throw runtime_error( "expat: unable to create parser" );

    tSAXDatabaseHandler handler( *this, progress, parser );

    XML_SetUserData( parser, &handler );
    XML_SetElementHandler( parser, startExpatElement, endExpatElement );

    char buffer[1<<8];
    while ( !dbfile.atEnd() )
    {
      Q_LONG read_bytes = dbfile.readBlock( buffer, sizeof(buffer) );
      if ( XML_Parse( parser, buffer, read_bytes, dbfile.atEnd() ) == 0 )
      {
	throw runtime_error( "expat: xml parse error" );
	// *** FIXME XML_ErrorString( XML_GetErrorCode( parser ) ) );
      }
    }

    XML_ParserFree( parser );
    dbfile.close();
  }
  else
    runtimeError( qApp->translate( "ErrorMessages", "Database file not accessible: %1" ).arg( filename ) );
}




void tDatabase::save( const QString &filename, tProgress *progress )
{
  if ( progress )
  {
    progress->setTotalSteps( 5 );
    progress->setProgress( 0 );
  }

  QDomDocument configxml;
  QDomElement doc_element = configxml.createElement( "madmandb" );
  configxml.appendChild( doc_element );

  // save song collection
  doc_element.appendChild( SongCollection.serialize( configxml ) );
  if (progress ) progress->setProgress( 1 );

  // save directories
  QDomElement dirs_element = configxml.createElement( "directories" );
  FOREACH( first_d, DirectoryList, tDirectoryList )
  {
    QDomElement dir_element = configxml.createElement( "directory" );
    dir_element.setAttribute( "name", *first_d );
    dirs_element.appendChild( dir_element );
  }
  doc_element.appendChild( dirs_element );
  if ( progress ) progress->setProgress( 2 );

  // order matters, song set tree needs to be serialized after songs
  if ( SongSetTree )
    doc_element.appendChild( SongSetTree->serialize( configxml ) );
  if ( progress ) progress->setProgress( 3 );

  doc_element.appendChild( History.serialize( configxml ) );
  if ( progress ) progress->setProgress( 4 );

  QFile dbfile( filename );
  if ( dbfile.open( IO_WriteOnly ) )
  {
    QTextStream stream( &dbfile );
    stream << configxml;
  }
  else
    runtimeError( qApp->translate( "ErrorMessages", "Database file not writable: %1" ).arg( filename ) );
  if ( progress ) progress->setProgress( 5 );
}
