#include <qmessagebox.h>
#include <stdexcept>
#include <qinputdialog.h>
#include <qregexp.h>
#include <qapplication.h>

#include "database/song_set.h"
#include "utility/scanner.h"
#include "utility/progress.h"

#include "song_set.moc"




namespace
{
  template<class T>
  QDomNode serializeUniqueIdContainer( QDomDocument &doc, QString const &name, T const &uid_list )
  {
    QDomElement result = doc.createElement( name );
    FOREACH_CONST( first, uid_list, typename T )
    {
      QDomElement song = doc.createElement( "song" );
      result.appendChild( song );
      song.setAttribute( "unique_id", *first );
    }
    return result;
  }




  const int TT_CRIT_PAREN_OPEN = 0;
  const int TT_CRIT_PAREN_CLOSE = 1;
  const int TT_CRIT_NOT = 2;
  const int TT_CRIT_AND = 3;
  const int TT_CRIT_OR = 4;

  const int TT_CRIT_ALBUM_CRIT = 100;
  const int TT_CRIT_ARTIST_CRIT = 101;
  const int TT_CRIT_PERFORMER_CRIT = 102;
  const int TT_CRIT_FILENAME_CRIT = 103;
  const int TT_CRIT_TITLE_CRIT = 104;
  const int TT_CRIT_GENRE_CRIT = 105;
  const int TT_CRIT_ANY_CRIT = 106;
  const int TT_CRIT_ALL_CRIT = 107;
  const int TT_CRIT_NONE_CRIT = 108;
  const int TT_CRIT_RATING_CRIT = 109;
  const int TT_CRIT_PLAYCOUNT_CRIT = 110;
  const int TT_CRIT_EXISTEDFORDAYS_CRIT = 111;
  const int TT_CRIT_FULLPLAYRATIO_CRIT = 112;

  const int TT_CRIT_COMPLETE_MATCH = 200;
  const int TT_CRIT_SUBSTRING_MATCH = 201;
  const int TT_CRIT_RE_MATCH = 202;
  const int TT_CRIT_FUZZY_MATCH = 203;

  const int TT_CRIT_NUM_LESS_MATCH = 204;
  const int TT_CRIT_NUM_LESSEQUAL_MATCH = 205;
  const int TT_CRIT_NUM_EQUAL_MATCH = 206;
  const int TT_CRIT_NUM_GREATER_MATCH = 207;
  const int TT_CRIT_NUM_GREATEREQUAL_MATCH = 208;

  const int TT_CRIT_STRING = 300;
  const int TT_CRIT_QUOTED_STRING = 301;



  class tCriterionScanner : public tScanner
  {
    public:
      tCriterionScanner()
      {
	addTokenDescriptor( TT_CRIT_PAREN_OPEN, "\\(" );
	addTokenDescriptor( TT_CRIT_PAREN_CLOSE, "\\)" );
	addTokenDescriptor( TT_CRIT_NOT, "\\!" );
	addTokenDescriptor( TT_CRIT_AND, "\\&" );
	addTokenDescriptor( TT_CRIT_OR, "\\|" );
	addTokenDescriptor( TT_CRIT_ALBUM_CRIT, "\\~album" );
	addTokenDescriptor( TT_CRIT_ARTIST_CRIT, "\\~artist" );
	addTokenDescriptor( TT_CRIT_PERFORMER_CRIT, "\\~performer" );
	addTokenDescriptor( TT_CRIT_FILENAME_CRIT, "\\~filename" );
	addTokenDescriptor( TT_CRIT_TITLE_CRIT, "\\~title" );
	addTokenDescriptor( TT_CRIT_GENRE_CRIT, "\\~genre" );
	addTokenDescriptor( TT_CRIT_ANY_CRIT, "\\~any" );
	addTokenDescriptor( TT_CRIT_RATING_CRIT, "\\~rating" );
	addTokenDescriptor( TT_CRIT_PLAYCOUNT_CRIT, "\\~playcount" );
	addTokenDescriptor( TT_CRIT_EXISTEDFORDAYS_CRIT, "\\~existedfordays" );
	addTokenDescriptor( TT_CRIT_FULLPLAYRATIO_CRIT, "\\~fullplayratio" );
	addTokenDescriptor( TT_CRIT_ALL_CRIT, "\\~all(\b|$)" );
	addTokenDescriptor( TT_CRIT_NONE_CRIT, "\\~none(\b|$)" );

	addTokenDescriptor( TT_CRIT_COMPLETE_MATCH, "complete\\:" );
	addTokenDescriptor( TT_CRIT_SUBSTRING_MATCH, "substring\\:" );
	addTokenDescriptor( TT_CRIT_RE_MATCH, "re\\:" );
	addTokenDescriptor( TT_CRIT_FUZZY_MATCH, "fuzzy\\:" );

	addTokenDescriptor( TT_CRIT_NUM_LESSEQUAL_MATCH, "\\<\\=" );
	addTokenDescriptor( TT_CRIT_NUM_GREATEREQUAL_MATCH, "\\>\\=" );
	addTokenDescriptor( TT_CRIT_NUM_LESS_MATCH, "\\<" );
	addTokenDescriptor( TT_CRIT_NUM_EQUAL_MATCH, "\\=" );
	addTokenDescriptor( TT_CRIT_NUM_GREATER_MATCH, "\\>" );

	addTokenDescriptor( TT_CRIT_FUZZY_MATCH, "fuzzy\\:" );
	addTokenDescriptor( TT_CRIT_STRING, "(\\w|[.,/\\\\])+" );
	addTokenDescriptor( TT_CRIT_QUOTED_STRING, "\\\"([^\"]|\\\\\")*\\\"" );
	addTokenDescriptor( 0, "[ \n\r\t]+", true, true );
	addTokenDescriptor( 0, "\\{[^}]*\\}", true, true );
      }
  } CriterionScanner;
}





QString quoteString( QString victim )
{
  return "\"" + victim.replace( QRegExp( "\\\"" ), "\\\"" ) + "\"";
}




// criteria -------------------------------------------------------------------
namespace
{
  class tTextCriterion 
  {
    public:
      virtual ~tTextCriterion()
      { }
      virtual float match( const QString &str ) = 0;
      virtual bool isBinary() = 0;
  };

  class tNumericCriterion 
  {
    public:
      virtual ~tNumericCriterion()
      { }
      virtual float match( double x ) = 0;
      virtual bool isBinary() = 0;
  };






// text criteria --------------------------------------------------------------
  class tFuzzyCriterion : public tTextCriterion
  {
      typedef hash_map<QString, int, hash_QString> tnGramTable;

      tnGramTable		SmallNGramTable;
      tnGramTable		BigNGramTable;
      QString		Pattern;
      int			PatternLength;

      static int const SmallN = 3;
      static int const BigN = 5;

    public:
      tFuzzyCriterion( const QString &op );
      float match( const QString &str );
      bool isBinary() 
      { 
	return false; 
      }
  };




  // only for odd n
#define FOREACH_N_GRAM( STRING, N, OPERATION ) \
    { \
      int strlength = STRING.length(); \
      int nhalf = (N) / 2; \
      { \
	QString ngram = " " + STRING.mid( 0, N - 1 ); \
	OPERATION; \
      } \
      int i = nhalf; \
      while ( i < strlength - nhalf ) \
      { \
	QString ngram = STRING.mid( i - nhalf, N );\
	OPERATION; \
	if ( STRING[ i + nhalf ].isSpace() ) \
	  i += (N) - 1; \
	else \
	  i++; \
      } \
      { \
	QString ngram = STRING.mid( strlength - ( (N) - 1 ), (N) - 1 ) + " "; \
	OPERATION; \
      } \
    }




  tFuzzyCriterion::tFuzzyCriterion( const QString &op )
  {
    Pattern = op;
    PatternLength = op.length();

    FOREACH_N_GRAM( op, SmallN, 
	SmallNGramTable[ ngram ] = 1; );
    FOREACH_N_GRAM( op, BigN, 
	BigNGramTable[ ngram ] = 1; );
  }




  float tFuzzyCriterion::match( const QString &str )
  {
    if ( (int) str.length() < SmallN )
      return 0;

    int small_matches = 0, big_matches = 0;

    FOREACH_N_GRAM( str, SmallN, 
      if ( SmallNGramTable.find( ngram ) != SmallNGramTable.end() )
	small_matches++;
	);
    FOREACH_N_GRAM( str, BigN, 
      if ( BigNGramTable.find( ngram ) != BigNGramTable.end() )
	big_matches++; );

    float result = float( small_matches * SmallN + big_matches * BigN )/
      float( SmallNGramTable.size() * SmallN + BigNGramTable.size() * BigN );
    if ( result > 1 )
      return 1;
    else
      return result;
  }




  class tSubstringCriterion : public tTextCriterion
  {
    private:
      QString Operand;
    public:
      tSubstringCriterion( const QString &op )
      : Operand( op )
      { }
      float match( const QString &str )
      { return str.find( Operand, 0, true ) != -1 ? 1 : 0; }
      bool isBinary() 
      { 
	return true; 
      }
  };




  class tCompleteCriterion : public tTextCriterion
  {
    private:
      QString Operand;
    public:
      tCompleteCriterion( const QString &op )
      : Operand( op )
      { }
      float match( const QString &str )
      { return str == Operand ? 1 : 0; }
      bool isBinary() 
      { 
	return true; 
      }
  };




  class tRegexCriterion : public tTextCriterion {
    private:
      QRegExp Operand;
    public:
      tRegexCriterion( const QString &op )
      : Operand( op )
      { 
      }
      float match( const QString &str )
      { return Operand.search( str ) ? 1 : 0; }
      bool isBinary() 
      { 
	return true; 
      }
  };




  // numeric criteria -----------------------------------------------------------
  class tRelationalNumericCriterion : public tNumericCriterion {
    private:
      float	ReferenceNumber;
      int		TokenType;

    public:
      tRelationalNumericCriterion( double reference, int token_type )
      : ReferenceNumber( reference ), TokenType( token_type )
      { }
      float match( double x )
      {
	switch ( TokenType )
	{
	  case TT_CRIT_NUM_LESS_MATCH:
	    return x < ReferenceNumber ? 1 : 0;
	  case TT_CRIT_NUM_LESSEQUAL_MATCH:
	    return x <= ReferenceNumber ? 1 : 0;
	  case TT_CRIT_NUM_EQUAL_MATCH:
	    return x == ReferenceNumber ? 1 : 0;
	  case TT_CRIT_NUM_GREATER_MATCH:
	    return x > ReferenceNumber ? 1 : 0;
	  case TT_CRIT_NUM_GREATEREQUAL_MATCH:
	    return x >= ReferenceNumber ? 1 : 0;
	  default:
	    return 0;
	}      
      }
      bool isBinary() 
      { 
	return true; 
      }
  };




  // criteria -------------------------------------------------------------------
  class tAnyCriterion : public tCriterion
  {
      auto_ptr<tTextCriterion>		TextCriterion;
    public:
      tAnyCriterion( tTextCriterion *tc )
      : TextCriterion( tc ) 
      { }
      float operator() ( tSong *song ) const
      {
	return TextCriterion->match( song->key() );
      }
      bool isBinary() 
      { 
	return TextCriterion->isBinary(); 
      }
  };




  class tArtistCriterion : public tCriterion
  {
      auto_ptr<tTextCriterion>		TextCriterion;
    public:
      tArtistCriterion( tTextCriterion *tc )
      : TextCriterion( tc ) 
      { }
      float operator() ( tSong *song ) const
      {
	return TextCriterion->match( song->artist().lower() );
      }
      bool isBinary() 
      { 
	return TextCriterion->isBinary(); 
      }
  };




  class tPerformerCriterion : public tCriterion
  {
      auto_ptr<tTextCriterion>		TextCriterion;
    public:
      tPerformerCriterion( tTextCriterion *tc )
      : TextCriterion( tc ) 
      { }
      float operator() ( tSong *song ) const
      {
	return TextCriterion->match( song->performer().lower() );
      }
      bool isBinary() 
      { 
	return TextCriterion->isBinary(); 
      }
  };




  class tAlbumCriterion : public tCriterion
  {
      auto_ptr<tTextCriterion>		TextCriterion;
    public:
      tAlbumCriterion( tTextCriterion *tc )
      : TextCriterion( tc ) 
      { }
      float operator() ( tSong *song ) const
      {
	return TextCriterion->match( song->album().lower() );
      }
      bool isBinary() 
      { 
	return TextCriterion->isBinary(); 
      }
  };




  class tTitleCriterion : public tCriterion
  {
      auto_ptr<tTextCriterion>		TextCriterion;
    public:
      tTitleCriterion( tTextCriterion *tc )
      : TextCriterion( tc ) 
      { }
      float operator() ( tSong *song ) const
      {
	return TextCriterion->match( song->title().lower() );
      }
      bool isBinary() 
      { 
	return TextCriterion->isBinary(); 
      }
  };




  class tGenreCriterion : public tCriterion
  {
      auto_ptr<tTextCriterion>		TextCriterion;
    public:
      tGenreCriterion( tTextCriterion *tc )
      : TextCriterion( tc ) 
      { }
      float operator() ( tSong *song ) const
      {
	return TextCriterion->match( song->genre().lower() );
      }
      bool isBinary() 
      { 
	return TextCriterion->isBinary(); 
      }
  };




  class tFilenameCriterion : public tCriterion
  {
      auto_ptr<tTextCriterion>		TextCriterion;
    public:
      tFilenameCriterion( tTextCriterion *tc )
      : TextCriterion( tc ) 
      { }
      float operator() ( tSong *song ) const
      {
	return TextCriterion->match( song->filename().lower() );
      }
      bool isBinary() 
      { 
	return TextCriterion->isBinary(); 
      }
  };




  class tRatingCriterion : public tCriterion
  {
      auto_ptr<tNumericCriterion>		NumericCriterion;
    public:
      tRatingCriterion( tNumericCriterion *nc )
      : NumericCriterion( nc ) 
      { }
      float operator() ( tSong *song ) const
      {
	return NumericCriterion->match( song->rating() );
      }
      bool isBinary() 
      { 
	return NumericCriterion->isBinary(); 
      }
  };




  class tPlayCountCriterion : public tCriterion
  {
      auto_ptr<tNumericCriterion>		NumericCriterion;
    public:
      tPlayCountCriterion( tNumericCriterion *nc )
      : NumericCriterion( nc ) 
      { }
      float operator() ( tSong *song ) const
      {
	return NumericCriterion->match( song->playCount() );
      }
      bool isBinary() 
      { 
	return NumericCriterion->isBinary(); 
      }
  };




  class tExistedForDaysCriterion : public tCriterion
  {
      auto_ptr<tNumericCriterion>		NumericCriterion;
    public:
      tExistedForDaysCriterion( tNumericCriterion *nc )
      : NumericCriterion( nc ) 
      { }
      float operator() ( tSong *song ) const
      {
	time_t now = time( NULL );
	time_t song_added = song->existsSince();
	double diff_in_days = (double) (now - song_added) / (3600. * 24);

	return NumericCriterion->match( diff_in_days );
      }
      bool isBinary() 
      { 
	return NumericCriterion->isBinary(); 
      }
  };




  class tFullPlayRatioCriterion : public tCriterion
  {
      auto_ptr<tNumericCriterion>		NumericCriterion;
    public:
      tFullPlayRatioCriterion( tNumericCriterion *nc )
      : NumericCriterion( nc ) 
      { }
      float operator() ( tSong *song ) const
      {
	if ( song->playCount() == 0 )
	  return false;
	double ratio = (double) song->fullPlayCount() / (double ) song->playCount();
	return NumericCriterion->match( ratio );
      }
      bool isBinary() 
      { 
	return NumericCriterion->isBinary(); 
      }
  };




  class tOrCriterion : public tCriterion
  {
      auto_ptr<tCriterion>	Operand1,Operand2;
    public:
      tOrCriterion( tCriterion *op1, tCriterion *op2 )
	: Operand1( op1 ), Operand2( op2 )
      { }
      float operator() ( tSong *song ) const
      {
	return max( (*Operand1)( song ), (*Operand2)( song ) );
      }
      bool isBinary() 
      { 
	return Operand1->isBinary() && Operand2->isBinary(); 
      }
  };




  class tAndCriterion : public tCriterion
  {
      auto_ptr<tCriterion>	Operand1,Operand2;
    public:
      tAndCriterion( tCriterion *op1, tCriterion *op2 )
	: Operand1( op1 ), Operand2( op2 )
      { }
      float operator() ( tSong *song ) const
      {
	return min( (*Operand1)( song ), (*Operand2)( song ) );
      }
      bool isBinary() 
      { 
	return Operand1->isBinary() && Operand2->isBinary(); 
      }
  };




  class tNotCriterion : public tCriterion
  {
      auto_ptr<tCriterion>	Operand;
    public:
      tNotCriterion( tCriterion *op )
	: Operand( op )
      { }
      float operator() ( tSong *song ) const
      {
	return 1 - (*Operand)( song );
      }
      bool isBinary() 
      { 
	return Operand->isBinary(); 
      }
  };




  class tTrueCriterion : public tCriterion
  {
    public:
      float operator() ( tSong *song ) const
      {
	return 1;
      }
      bool isBinary() 
      { 
	return true; 
      }
  };




  class tFalseCriterion : public tCriterion
  {
    public:
      float operator() ( tSong *song ) const
      {
	return false;
      }
      bool isBinary() 
      { 
	return true; 
      }
  };
}




// parsing --------------------------------------------------------------------
namespace
{
#define ADVANCE \
  first++; \
  if (first == last) \
    throw runtime_error( "unexpected end of expression" );
#define EXPECT(WHAT,STRINGIFIED) \
  if (first == last) \
    throw runtime_error( "unexpected end of expression" ); \
  if (first->TokenType != WHAT) \
    runtimeError( qApp->translate( "ErrorMessages", "Expected '%1', found '%2'." ).arg( STRINGIFIED ).arg( first->Text ) );





  QString getText( tTokenList::iterator &first, tTokenList::iterator last )
  {
    if ( first->TokenType == TT_CRIT_STRING )
      return (first++)->Text;
    else if ( first->TokenType == TT_CRIT_QUOTED_STRING )
    {
      QString result = (first++)->Text;
      result = result.mid( 1, result.length() - 2 );
      // unquote string
      result.replace( QRegExp( "\\\\\\\"" ), "\"");
      return result;
    }
    else
      runtimeError( qApp->translate( "ErrorMessages", "Expected string expression, found: '%1'." ).arg( first->Text ) );
  }




  tNumericCriterion *parseNumericCriterion( tTokenList::iterator &first, tTokenList::iterator last )
  {
    double ref = 0;
    int token_type = TT_CRIT_NUM_EQUAL_MATCH; 
    switch ( first->TokenType )
    {
      case TT_CRIT_NUM_LESS_MATCH:
      case TT_CRIT_NUM_LESSEQUAL_MATCH:
      case TT_CRIT_NUM_EQUAL_MATCH:
      case TT_CRIT_NUM_GREATER_MATCH:
      case TT_CRIT_NUM_GREATEREQUAL_MATCH:
	token_type = first->TokenType;
	ADVANCE;
      default:
	  bool ok;
	  QString ref_num_string = getText( first, last );
	  ref = ref_num_string.toDouble( &ok );
	  if ( !ok )
	    runtimeError( qApp->translate( "ErrorMessages", "Expected number, found: %1") .arg( ref_num_string ) );
    }
    return new tRelationalNumericCriterion( ref, token_type );
  }




  tTextCriterion *parseTextCriterion( tTokenList::iterator &first, tTokenList::iterator last, bool use_fuzzy )
  {
    if ( first->TokenType == TT_CRIT_COMPLETE_MATCH )
    {
      ADVANCE;
      return new tCompleteCriterion( getText( first, last ) );
    }
    if ( first->TokenType == TT_CRIT_SUBSTRING_MATCH )
    {
      ADVANCE;
      return new tSubstringCriterion( getText( first, last ) );
    }
    if ( first->TokenType == TT_CRIT_RE_MATCH )
    {
      ADVANCE;
      return new tRegexCriterion( getText( first, last ) );
    }
    if ( first->TokenType == TT_CRIT_FUZZY_MATCH )
    {
      ADVANCE;
      return new tFuzzyCriterion( getText( first, last ) );
    }
    QString pattern = getText( first, last );
    if ( pattern.length() < 4 || !use_fuzzy )
      return new tSubstringCriterion( pattern );
    else
      return new tFuzzyCriterion( pattern );
  }




  tCriterion *parseAtomicCriterion( tTokenList::iterator &first, tTokenList::iterator last, bool use_fuzzy )
  {
#define PARSE_PAREN_TEXT_CRITERION( TOKENTYPE, CLASSNAME )\
    if ( first->TokenType == TOKENTYPE ) \
    { \
      ADVANCE; \
      EXPECT( TT_CRIT_PAREN_OPEN, "')'" ); \
      ADVANCE; \
      auto_ptr<tCriterion> result( new CLASSNAME( parseTextCriterion( first, last, use_fuzzy ) ) ); \
      EXPECT( TT_CRIT_PAREN_CLOSE, "')'" ); \
      first++; \
      return result.release(); \
    }
    PARSE_PAREN_TEXT_CRITERION( TT_CRIT_ALBUM_CRIT, tAlbumCriterion );
    PARSE_PAREN_TEXT_CRITERION( TT_CRIT_ARTIST_CRIT, tArtistCriterion );
    PARSE_PAREN_TEXT_CRITERION( TT_CRIT_PERFORMER_CRIT, tPerformerCriterion );
    PARSE_PAREN_TEXT_CRITERION( TT_CRIT_FILENAME_CRIT, tFilenameCriterion );
    PARSE_PAREN_TEXT_CRITERION( TT_CRIT_TITLE_CRIT, tTitleCriterion );
    PARSE_PAREN_TEXT_CRITERION( TT_CRIT_GENRE_CRIT, tGenreCriterion );
    PARSE_PAREN_TEXT_CRITERION( TT_CRIT_ANY_CRIT, tAnyCriterion );

#define PARSE_PAREN_NUM_CRITERION( TOKENTYPE, CLASSNAME )\
    if ( first->TokenType == TOKENTYPE ) \
    { \
      ADVANCE; \
      EXPECT( TT_CRIT_PAREN_OPEN, "'('" ); \
      ADVANCE; \
      auto_ptr<tCriterion> result( new CLASSNAME( parseNumericCriterion( first, last ) ) ); \
      EXPECT( TT_CRIT_PAREN_CLOSE, "')'" ); \
      first++; \
      return result.release(); \
    }
    PARSE_PAREN_NUM_CRITERION( TT_CRIT_RATING_CRIT, tRatingCriterion );
    PARSE_PAREN_NUM_CRITERION( TT_CRIT_PLAYCOUNT_CRIT, tPlayCountCriterion );
    PARSE_PAREN_NUM_CRITERION( TT_CRIT_EXISTEDFORDAYS_CRIT, tExistedForDaysCriterion );
    PARSE_PAREN_NUM_CRITERION( TT_CRIT_FULLPLAYRATIO_CRIT, tFullPlayRatioCriterion );

    if ( first->TokenType == TT_CRIT_ALL_CRIT )
    {
      first++;
      return new tTrueCriterion;
    }
    if ( first->TokenType == TT_CRIT_NONE_CRIT )
    {
      first++;
      return new tFalseCriterion;
    }

    return new tAnyCriterion( parseTextCriterion( first, last, use_fuzzy ) );
  }




  const int PRECEDENCE_UNARY = 200;
  const int PRECEDENCE_AND = 150;
  const int PRECEDENCE_OR = 100;
  tCriterion *parseCriterion( tTokenList::iterator &first, tTokenList::iterator last, bool use_fuzzy, const int min_precedence = 0 )
  {
    QString ft = first->Text;
    auto_ptr<tCriterion> op1;
    if ( first->TokenType == TT_CRIT_PAREN_OPEN )
    {
      ADVANCE;
      auto_ptr<tCriterion> temp_result( parseCriterion( first, last, use_fuzzy ) );
      op1 = temp_result;
      EXPECT( TT_CRIT_PAREN_CLOSE, "')'" );
      first++;
    }
    else if ( first->TokenType == TT_CRIT_NOT && min_precedence <= PRECEDENCE_UNARY )
    {
      ADVANCE;
      auto_ptr<tCriterion> op( parseCriterion( first, last, use_fuzzy, PRECEDENCE_UNARY ) );
      auto_ptr<tCriterion> temp_result( new tNotCriterion( op.get() ) );
      op.release();
      op1 = temp_result;
    }
    else 
    {
      auto_ptr<tCriterion> temp_result( parseAtomicCriterion( first, last, use_fuzzy ) );
      op1 = temp_result;
    }
    
    bool made_progress = true;
    while ( first != last && made_progress )
    {
      made_progress = false;
      if ( first-> TokenType == TT_CRIT_AND && min_precedence <= PRECEDENCE_AND )
      {
	ADVANCE;
	auto_ptr<tCriterion> temp_result( 
	    new tAndCriterion( op1.get(), parseCriterion( first, last, use_fuzzy, PRECEDENCE_AND ) ) );
	op1.release();
	op1 = temp_result;
	made_progress = true;
      }
      else if ( first-> TokenType == TT_CRIT_OR && min_precedence <= PRECEDENCE_OR )
      {
	ADVANCE;
	auto_ptr<tCriterion> temp_result( 
	    new tOrCriterion( op1.get(), parseCriterion( first, last, use_fuzzy, PRECEDENCE_OR ) ) );
        op1.release();
	op1 = temp_result;
	made_progress = true;
      }
      else if ( min_precedence <= PRECEDENCE_AND )
      {
	try
	{
	  tTokenList::iterator myfirst = first;
	  auto_ptr<tCriterion> temp_result( 
	      new tAndCriterion( op1.get(), parseCriterion( myfirst, last, use_fuzzy, PRECEDENCE_AND ) ) );
	  op1.release();
	  op1 = temp_result;
	  made_progress = true;
	  first = myfirst;
	}
	catch (...)
	{
	}
      }
    }

    return op1.release();
  }
}




auto_ptr<tCriterion> parseCriterion( QString criterion, bool use_fuzzy )
{
  tTokenList tokens;
  CriterionScanner.scan( criterion, tokens );
  tTokenList::iterator first = tokens.begin(), last = tokens.end();

  if ( first == last )
    return auto_ptr<tCriterion>( new tTrueCriterion );
  auto_ptr<tCriterion> result( 
      parseCriterion( first, last, use_fuzzy ) );

  if ( first != last )
    runtimeError( qApp->translate( "ErrorMessages", "Unrecognized tokens at end of criterion" ) );

  return result;
}




// tSongSet -------------------------------------------------------------------
tSongSet::tSongSet()
    : SongCollection( NULL ), ParsedCriterion( NULL ),
    BulkChangeLevel( 0 ), MatchThreshold( 0.1 )
{
  setCriterion( "~none" );
}




void tSongSet::setSongCollection( tSongCollection *collection )
{
  if ( SongCollection )
    disconnect( SongCollection, NULL, this, NULL );

  Rendering.clear();
  RenderingSet.clear();
  PositiveSet.clear();
  NegativeSet.clear();
  
  SongCollection = collection;
  if ( SongCollection )
  {
    connect( SongCollection, SIGNAL( notifyBeginBulkChange() ),
	this, SLOT( beginBulkChange() ) );
    connect( SongCollection, SIGNAL( notifyEndBulkChange() ),
	this, SLOT( endBulkChange() ) );
    connect( SongCollection, SIGNAL( notifyAddition( tSong * ) ),
	this, SLOT( noticeCollectionAddition( tSong * ) ) );
    connect( SongCollection, SIGNAL( notifyDeletion( tSong * ) ),
	this, SLOT( noticeCollectionDeletion( tSong * ) ) );
    connect( SongCollection, SIGNAL( notifySongModified( tSong * ) ),
	this, SLOT( noticeSongModified( tSong * ) ) );

    reevaluateCriterion();
  }
}




tSongSet &tSongSet::operator=( const tSongSet &src )
{
  SongCollection = src.SongCollection;
  Criterion = src.Criterion;
  Rendering = src.Rendering;
  RenderingSet = src.RenderingSet;
  PositiveSet = src.PositiveSet;
  NegativeSet = src.NegativeSet;

  setCriterion( Criterion );
  hasChanged( true );
  return *this;
}




QDomNode tSongSet::serialize( QDomDocument &doc )
{
  QDomElement result = doc.createElement( "song_set" );
  result.setAttribute( "criterion", Criterion );

  tUniqueIdList rendering;
  render( rendering );

  result.appendChild( serializeUniqueIdContainer( doc, "rendering", rendering ) );
  result.appendChild( serializeUniqueIdContainer( doc, "positiveset", PositiveSet ) );
  result.appendChild( serializeUniqueIdContainer( doc, "negativeset", NegativeSet ) );

  return result;
}




void tSongSet::setRendering( const tUniqueIdList &list )
{
  FOREACH_CONST( first, list, tUniqueIdList )
  {
    tSong *song = SongCollection->getByUniqueId( *first );
    if ( song )
    {
      RenderingSet.insert( *first );
      tRenderingEntry entry;
      entry.Song = SongCollection->getByUniqueId( *first );
      entry.CriterionMatch = (*ParsedCriterion)( entry.Song );
      Rendering.push_back( entry );
    }
  }
}




void tSongSet::setPositiveSet( const tUniqueIdList &list )
{
  FOREACH_CONST( first, list, tUniqueIdList )
  {
    tSong *song = SongCollection->getByUniqueId( *first );
    if ( song )
      PositiveSet.insert( *first );
  }
}




void tSongSet::setNegativeSet( const tUniqueIdList &list )
{
  FOREACH_CONST( first, list, tUniqueIdList )
  {
    tSong *song = SongCollection->getByUniqueId( *first );
    if ( song )
      NegativeSet.insert( *first );
  }
}




void tSongSet::clear()
{
  Rendering.clear();
  RenderingSet.clear();
  PositiveSet.clear();
  NegativeSet.clear();
  setCriterion( "~none" );
}




void tSongSet::render( tSongList &songlist )
{
  songlist.clear();
  FOREACH_CONST( first, Rendering, tRendering )
    songlist.push_back( first->Song );
}




void tSongSet::render( tRendering &rendering)
{
  rendering = Rendering;
}




void tSongSet::render( tUniqueIdList &uidlist )
{
  uidlist.clear();
  FOREACH( first, Rendering, tRendering )
    uidlist.push_back( first->Song->uniqueId() );
}




void tSongSet::sortBy( tSongField field )
{
  sort( Rendering.begin(), Rendering.end(), less_rendering( field ) );
  hasChanged( false );
}




void tSongSet::jumble()
{
  srand( time( NULL ) );

  for ( unsigned i = 0; i < Rendering.size(); i++ )
  {
    int new_index = rand() % Rendering.size();
    tRenderingEntry buffer_old = Rendering[ new_index ];
    Rendering[ new_index ] = Rendering[ i ];
    Rendering[ i ] = buffer_old;
  }
  hasChanged( false );
}




void tSongSet::reverse()
{
  std::reverse( Rendering.begin(), Rendering.end() );
  hasChanged( false );
}




void tSongSet::add( tSong *song, int preferred_index )
{
  float match = (*ParsedCriterion)( song );

  addToRendering( song, match, preferred_index );
  if ( match < MatchThreshold )
    PositiveSet.insert( song->uniqueId() );
  NegativeSet.erase( song->uniqueId() );

  hasChanged( false );
}




void tSongSet::add( const tFilename &filename, int preferred_index )
{
  tSong *song = SongCollection->getByFilename( filename );
  if ( song )
    add( song, preferred_index );
  else
    runtimeError( tr( "Unable to add unknown file %1"  ).arg( filename ) );
}




void tSongSet::remove( int index )
{
  if ( index >= (int) Rendering.size() )
  {
    qWarning( "tried to remove song out of bounds" );
    return;
  }

  tSong *song = Rendering[ index ].Song;
  removeFromRendering( index );

  float match = (*ParsedCriterion)( song );

  PositiveSet.erase( song->uniqueId() );
  if ( match >= MatchThreshold )
    NegativeSet.insert( song->uniqueId() );

  hasChanged( false );
}




void tSongSet::remove( tSong *song )
{
  removeFromRendering( song );

  PositiveSet.erase( song->uniqueId() );
  float match = (*ParsedCriterion)( song );
  if ( match >= MatchThreshold )
    NegativeSet.insert( song->uniqueId() );

  hasChanged( false );
}




void tSongSet::setCriterion( const QString &new_crit, bool use_fuzzy )
{
  if ( new_crit == Criterion )
    return;

  auto_ptr<tCriterion> parsed_crit( parseCriterion( new_crit, use_fuzzy ) );
  ParsedCriterion = parsed_crit;

  Criterion = new_crit;
}




void tSongSet::reevaluateCriterion( tProgress *progress )
{
  if ( progress )
    progress->setWhat( tr( "Searching..." ) );
  // reevaluate current rendering
  unsigned index = 0;
  while ( index < Rendering.size() )
  {
    tSong *song = Rendering[ index ].Song;

    float match = (*ParsedCriterion)( song );

    if ( match < MatchThreshold && PositiveSet.count( song->uniqueId() ) == 0 )
      removeFromRendering( index );
    else
    {
      Rendering[ index ].CriterionMatch = match;
      index++;
    }

    if ( progress )
    {
      progress->processEvents();
      if ( progress->wasCancelled() )
	return;
    }
  }

  // reevaluate the whole collection
  FOREACH_CONST( first, *SongCollection, tSongCollection )
  {
    tUniqueId uid = ( *first )->uniqueId();

    float match = (*ParsedCriterion)( (*first) );
    if ( match >= MatchThreshold && (NegativeSet.count( uid ) == 0)
         && !isInRendering( uid ) )
      addToRendering( *first, match );

    if ( progress )
    {
      progress->processEvents();
      if ( progress->wasCancelled() )
	return;
    }
  }


  hasChanged( true, progress );
}




bool tSongSet::hasBinaryCriterion()
{
  return ParsedCriterion->isBinary();
}




void tSongSet::beginBulkChange()
{
  BulkChangeLevel++;
  SubstantialBulkChange = false;
}




void tSongSet::endBulkChange()
{
  /* The if is necessary since this happens during loading:
   * Newly loaded song sets catch the endBulkChange() after
   * loading finishes, but weren't loaded for the beginBulkChange()
   * when it started.
   */

  if ( BulkChangeLevel > 0 )
    BulkChangeLevel--;
  hasChanged( SubstantialBulkChange );
}




void tSongSet::noticeCollectionAddition( tSong *song )
{
  float match = (*ParsedCriterion)( song );
  if ( match >= MatchThreshold && NegativeSet.count( song->uniqueId() ) == 0 )
    addToRendering( song, match );

  hasChanged( false );
}




void tSongSet::noticeCollectionDeletion( tSong *song )
{
  int index = 0;
  while ( index < (int) Rendering.size() )
  {
    if ( Rendering[index].Song == song )
      removeFromRendering( index );
    else
      index++;
  }

  NegativeSet.erase( song->uniqueId() );
  PositiveSet.erase( song->uniqueId() );

  hasChanged( false );
}




void tSongSet::noticeSongModified( tSong *song )
{
  // matchesCriterion needs to be called first since it may actually recursively
  // end up calling this function!
  float match = (*ParsedCriterion)( song );
  bool matches_criterion = match >= MatchThreshold;
  bool in_rendering = isInRendering( song->uniqueId() );

  if ( in_rendering )
  {
    if ( !matches_criterion && PositiveSet.count( song->uniqueId() ) == 0 )
    {
      removeFromRendering( song );
      hasChanged( false );
    }
  }
  else
  {
    if ( matches_criterion && NegativeSet.count( song->uniqueId() ) == 0 )
    {
      addToRendering( song, match );
      hasChanged( false );
    }
  }

  emit notifySongModified( song );
}




void tSongSet::hasChanged( bool substantial_change, tProgress *progress )
{
  if ( BulkChangeLevel == 0 )
    emit notifyChange( substantial_change, progress );
  else
  {
    if ( substantial_change )
      SubstantialBulkChange = true;
  }
}





void tSongSet::addToRendering( tSong *song, float match, int preferred_index )
{
  tUniqueId uid = song->uniqueId();

  tRenderingEntry entry;
  entry.Song = song;
  entry.CriterionMatch = match;

  if ( preferred_index == -1 || preferred_index > (int) Rendering.size() )
    Rendering.push_back( entry );
  else
    Rendering.insert( Rendering.begin() + preferred_index, entry );

  // multiple insertions don't hurt
  RenderingSet.insert( uid );
}




void tSongSet::removeFromRendering( tSong *song )
{
  int index = 0;
  while ( index < (int) Rendering.size() )
  {
    if ( Rendering[index].Song == song )
      removeFromRendering( index );
    else
      index++;
  }
}




void tSongSet::removeFromRendering( int index )
{
  tSong *song = Rendering[ index ].Song;
  tUniqueId uid = song->uniqueId();

  Rendering.erase( Rendering.begin() + index );

  // is there no other instance? 
  FOREACH( first, Rendering, tRendering )
  {
    if ( first->Song == song ) 
      return;
  }
  // then erase song from the rendering set
  RenderingSet.erase( uid );
}
