#include <pwd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <fstream>
#include <stdexcept>

#include <qtextbrowser.h>
#include <qheader.h>
#include <qprogressdialog.h>
#include <qdragobject.h>
#include <qapplication.h>
#include <qstatusbar.h>
#include <qfiledialog.h>
#include <qpushbutton.h>
#include <qlistbox.h>
#include <qmessagebox.h>
#include <qcombobox.h>
#include <qlineedit.h>
#include <qradiobutton.h>
#include <qfile.h>
#include <qstring.h>
#include <qaction.h>
#include <qsplitter.h>

#include "mainwin.h"
#include "utility/player.h"
#include "utility/progress.h"
#include "database/song_set.h"
#include "designer/helpbrowser_ui.h"
#include "ui/set_list_view.h"
#include "ui/multiline.h"
#include "ui/overview.h"

#include "mainwin.moc"




// tools ----------------------------------------------------------------------
namespace {
  class tCancellableSimpleProgress : public tProgress
  {
      bool Cancelled;
      int Counter;

    public:
      tCancellableSimpleProgress()
	: Cancelled( false ), Counter( 0 )
      {
      }
      void processEvents() 
      { 
	if ( ++Counter == 10 )
	{
	  Counter = 0;
	  qApp->processEvents();
	}
      }
      bool wasCancelled()
      {
	return Cancelled;
      }
      void cancel()
      {
	Cancelled = true;
      }
  };

  class tCancellableStatusBarProgress : public tCancellableSimpleProgress
  {
      QStatusBar *StatusBar;
    public:
      tCancellableStatusBarProgress( QStatusBar *bar )
	: StatusBar( bar )
      {
      }
      ~tCancellableStatusBarProgress()
      {
	StatusBar->clear();
      }
      void setWhat( const QString &what ) 
      { 
	StatusBar->message( what );
      }
  };

  class tProgressDialog : public tProgress
  {
      auto_ptr<QProgressDialog> Progress;
    public:
      tProgressDialog( tMainWindow *mwin, bool cancellable )
	: Progress( new QProgressDialog( mwin, "progress", true ) )
      {
	if ( !cancellable )
	  Progress->setCancelButtonText( QString::null );

	Progress->setAutoReset( false );
	Progress->setAutoClose( false );
	Progress->setMinimumDuration( 0 );
	Progress->setProgress( 0 );
      }
      bool wasCancelled()
      { 
	return Progress->wasCancelled();
      }
      void processEvents() 
      { 
	qApp->processEvents(); 
      }
      int progress() 
      { 
	return Progress->progress();
      }
      void setProgress( int p ) 
      { 
	Progress->setProgress( p );
      }
      void setTotalSteps( int p )
      { 
	Progress->setTotalSteps( p );
      }
      void setWhat( const QString &what ) 
      { 
	Progress->setLabelText( what );
      }
  };
}




// tool functions -------------------------------------------------------------
QString getConfigFile()
{
  passwd * pwd = getpwuid( getuid() );
  if ( pwd->pw_dir )
    return QString( pwd->pw_dir ) + "/.madmandb";
  else
    return ".madmandb";
}




// tMainWindow ----------------------------------------------------------------
tMainWindow::tMainWindow()
    : AllSongsView( *lstAllSongs, *this, "/madman/allsongsview" ),
    SongSetEditor( *lstSetEditor, *this, "/madman/songseteditor" ),
    CurrentSearchProgress( NULL )
{
  // set up gui 
  lstOverview->setSorting( -1 );
  lstOverview->header()->hide();
  lstSets->header()->hide();

  lstSets->setAcceptDrops( true );
  lstSets->viewport()->setAcceptDrops( true );
  
  connect( btnNewSet, SIGNAL( clicked( ) ), this, SLOT( addSongSet( ) ) );
  connect( btnCopySet, SIGNAL( clicked( ) ), this, SLOT( duplicateSongSet( ) ) );
  connect( btnRemoveSet, SIGNAL( clicked( ) ), this, SLOT( removeSongSet( ) ) );
  connect( btnEditSetCriterion, SIGNAL( clicked( ) ), this, SLOT( editMultilineSongSetCriterion( ) ) );

  connect( fileNewAction, SIGNAL( activated() ), this, SLOT( fileNew() ) );
  connect( fileOpenAction, SIGNAL( activated() ), this, SLOT( fileOpen() ) );
  connect( fileSaveAction, SIGNAL( activated() ), this, SLOT( fileSave() ) );
  connect( fileSaveAsAction, SIGNAL( activated() ), this, SLOT( fileSaveAs() ) );
  connect( filePreferencesAction, SIGNAL( activated() ), this, SLOT( filePreferences() ) );
  connect( fileExitAction, SIGNAL( activated() ), qApp, SLOT( closeAllWindows() ) );
  connect( rescanAction, SIGNAL( activated() ), this, SLOT( rescan() ) );
  connect( rereadTagsAction, SIGNAL( activated() ), 
      this, SLOT( rereadTags() ) );
  connect( updateOverviewAction, SIGNAL( activated() ), this, SLOT( buildOverviewTree() ) );
  connect( helpAboutAction, SIGNAL( activated() ), this, SLOT( helpAbout() ) );
  connect( helpAction, SIGNAL( activated() ), this, SLOT( help() ) );

  connect( btnHelpAll, SIGNAL( clicked() ), this, SLOT( help() ) );
  connect( btnHelpSet, SIGNAL( clicked() ), this, SLOT( help() ) );

  connect( lstSets, SIGNAL( itemRenamed( QListViewItem *, int, const QString & ) ),
           this, SLOT( renameSongSet( QListViewItem *, int, const QString & ) ) );
  connect( lstSets, SIGNAL( selectionChanged() ),
           this, SLOT( songSetSelectionChanged() ) );
  connect( editSongSetCriterion, SIGNAL( returnPressed() ),
           this, SLOT( songSetCriterionChanged() ) );
  connect( btnEditSetCriterion, SIGNAL( clicked() ),
           this, SLOT( songSetCriterionChanged() ) );
  connect( btnUpdateSetCriterion, SIGNAL( clicked() ),
           this, SLOT( songSetCriterionChanged() ) );

  connect( &AllSongsView, SIGNAL( add ( const tSongList & ) ),
             this, SLOT( add ( const tSongList & ) ) );
  connect( editSearch, SIGNAL( textChanged( const QString & ) ),
           this, SLOT( searchChanged( const QString & ) ) );

  connect( lstOverview, SIGNAL( selectionChanged( QListViewItem * ) ),
           this, SLOT( overviewSelectionChanged( QListViewItem * ) ) );

  // load and realize prefs
  Preferences.load( Settings );
  setupListView( AllSongsView, false );
  setupListView( SongSetEditor, false );

  loadSplitterAppearance( splitter1, Settings, "/madman/splitters/splitter1.state" );
  loadSplitterAppearance( splitter2, Settings, "/madman/splitters/splitter2.state" );
  loadSplitterAppearance( splitter3, Settings, "/madman/splitters/splitter3.state" );

  // load everything
  try
  {
    auto_ptr<tDatabase> new_db( new tDatabase );

    CurrentFilename = Settings.readEntry( "/madman/startup_file", QString::null, &FilenameValid );
    if ( FilenameValid )
    {
      auto_ptr<tProgressDialog> progress( new tProgressDialog( this, false ) );
      progress->setWhat( tr( "Loading database..." ) );
      new_db->load( CurrentFilename, progress.get() );
    }
    else
      throw runtime_error( QString2string( tr( "No startup file found" ) ) );

    if ( Preferences.ScanAtStartup )
      rescan( new_db.get() );

    setDatabase( new_db.release() );
  }
  catch (runtime_error &ex)
  {
    if ( QString( ex.what() ) != tr( "No startup file found" ) )
    {
      QMessageBox::information( this, tr( "madman" ),
	  tr( "Couldn't restore previously opened file:\n%1\nCreating a new file." ).
	  arg( QString( ex.what() ) ),
	  QMessageBox::Ok );
    }
    else
    {
      QMessageBox::information( this, tr( "Welcome to madman." ),
	  tr( "Welcome to madman.\n\n"
	  "We sincerely hope that you're about to discover a powerful new way\n"
	  "to enjoy your music. madman is a powerful tool, we advise you to take\n"
	  "your time to get familiar with its many features - the madman website at\n"
	  "http://madman.sf.net has lots of useful hints.\n\n"
	  "First, you might want to tell madman where you keep your music." ), 
	  QMessageBox::Ok );
    }

    FilenameValid = false;
    auto_ptr<tDatabase> new_db( new tDatabase );
    new_db->startNew();

    editPreferences( this, Preferences, new_db->DirectoryList, Settings, 1 );

    rescan( new_db.get() );

    setDatabase( new_db.release() );
  }

  // restore main window geometry ---------------------------------------------
  if ( Preferences.RememberGeometry )
  {
    {
      QString temp = Settings.readEntry( "/madman/main_window_position" );
      if ( !temp.isNull() )
      {
	int x,y;
	QTextIStream is( &temp );
	is >> x >> y;
	QPoint p( x,y );
	move( p );
      }
    }

    {
      QString temp = Settings.readEntry( "/madman/main_window_size" );
      if ( !temp.isNull() )
      {
	int x,y;
	QTextIStream is( &temp );
	is >> x >> y;
	QSize s( x,y );
	resize( s );
      }
    }
  }

  
  // set up song set tree -----------------------------------------------------
  lstSets->setAcceptDrops( true );
  lstSets->viewport()->setAcceptDrops( true );

  // set up lists -------------------------------------------------------------
  lstSets->setSorting( -1 );

  AllSongsSet.setCriterion( ""  );
  AllSongsSet.reevaluateCriterion();

  AllSongsView.setSongSet( &AllSongsSet );

  AllSongsView.setup();
  SongSetEditor.setup();

  updateAll();

  connect( &CurrentSongCheckerTimer, SIGNAL( timeout() ),
      this, SLOT( checkCurrentlyPlayingSong() ) );
  CurrentSongCheckerTimer.start( 500 );
}




tMainWindow::~tMainWindow()
{
  QPoint p = pos();
  QSize s = size();

  {
    QString temp;
    {
      QTextOStream os( &temp );
      os << p.x() << ' ' << p.y();
    }
    Settings.writeEntry( "/madman/main_window_position", temp );
  }

  {
    QString temp;
    {
      QTextOStream os( &temp );
      os << s.width() << ' ' << s.height();
    }
    Settings.writeEntry( "/madman/main_window_size", temp );
  }

  Preferences.save( Settings );

  AllSongsView.saveListViewAppearance( Settings );
  SongSetEditor.saveListViewAppearance( Settings );
  saveSplitterAppearance( splitter1, Settings, "/madman/splitters/splitter1.state" );
  saveSplitterAppearance( splitter2, Settings, "/madman/splitters/splitter2.state" );
  saveSplitterAppearance( splitter3, Settings, "/madman/splitters/splitter3.state" );
}




void tMainWindow::setDatabase( tDatabase *db )
{
  auto_ptr<tDatabase> new_db( db );
  Database = new_db;
  AllSongsSet.setSongCollection( &Database->SongCollection );
}




void tMainWindow::closeEvent( QCloseEvent *event )
{
  if ( fileSaveWithResult() )
  {
    Settings.writeEntry( "/madman/startup_file", CurrentFilename );
    tMainWindowBase::closeEvent( event );
  }
  else
  {
    if ( QMessageBox::warning( this, tr( "madman" ),
	  tr( "Failed to save your database. Do you still want to quit?" ),
	  QMessageBox::Yes, QMessageBox::No ) == QMessageBox::Yes )
      tMainWindowBase::closeEvent( event );
    else
      event->ignore();
  }
}




void tMainWindow::setupListView( tSongSetViewManager &mgr, bool save_before )
{
  if ( save_before )
    mgr.saveListViewAppearance( Settings );

  mgr.setupListView( Preferences.ColumnVisibility );

  mgr.loadListViewAppearance( Settings );
}




void tMainWindow::updateWindowCaption()
{
  if ( FilenameValid )
  {
    QFileInfo info( CurrentFilename );
    setCaption( tr( "madman - %1").arg( info.fileName() ) );
  }
  else
    setCaption( tr( "madman - <unnamed>" ) );
}




void tMainWindow::updateAll()
{
  updateSongSetTree( false, false );
  songSetSelectionChanged();
  buildOverviewTree();
  updateWindowCaption();
}




// slots ----------------------------------------------------------------------
void tMainWindow::rescan()
{
  rescan( Database.get() );
}




void tMainWindow::rescan( tDatabase *db )
{
  auto_ptr<tProgressDialog> progress( new tProgressDialog( this, false ) );
  db->SongCollection.scan( db->DirectoryList, progress.get() );
}




void tMainWindow::rereadTags()
{
  auto_ptr<tProgressDialog> progress( new tProgressDialog( this, true ) );
  Database->SongCollection.rereadTags( progress.get() );
}




void tMainWindow::fileNew()
{
  if ( QMessageBox::warning( this, tr( "madman" ),
	tr( "Are you sure you want to start over from scratch?" ),
	QMessageBox::Yes, QMessageBox::No ) == QMessageBox::Yes )
  {
    auto_ptr<tDatabase> new_db( new tDatabase );
    SongSetEditor.setSongSet( NULL );

    new_db->startNew();
    FilenameValid = false;

    editPreferences( this, Preferences, new_db->DirectoryList, Settings, 1 );
    rescan( new_db.get() );

    setDatabase( new_db.release() );
    updateAll();
  }
}




void tMainWindow::fileOpen()
{
  try
  {
    QString filename = QFileDialog::getOpenFileName( QString::null,
	tr( "madman Database (*.mad)" ),
	this );
    if ( filename.isNull() )
      return;

    SongSetEditor.setSongSet( NULL );

    auto_ptr<tDatabase> new_db( new tDatabase );

    auto_ptr<tProgressDialog> progress( new tProgressDialog( this, false ) );
    progress->setWhat( tr( "Loading database..." ) );
    new_db->load( filename, progress.get() );

    CurrentFilename = filename;
    FilenameValid = true;
    setDatabase( new_db.release() );

    updateAll();
  }
  catch ( exception &ex )
  {
    QMessageBox::warning( this, tr( "madman" ),
	tr( "Error opening the file '%1'" ).arg( string2QString( ex.what() ) ),
	QMessageBox::Ok, 0 );
  }
}




void tMainWindow::saveDatabase( const QString &name )
{
  QFileInfo file_info( name );
  QString dir_name = file_info.dirPath();
  QString trunk = file_info.baseName();
  QString ext = file_info.extension();

  if ( Preferences.BackupCount )
  {
    QDir my_dir( dir_name );
    my_dir.remove( ( trunk+"-backup-%1" ).arg( Preferences.BackupCount ) + "." + ext );

    for ( int dest_i = Preferences.BackupCount; dest_i > 1; dest_i-- )
    {
      QString srcname =( trunk+"-backup-%1" ).arg( dest_i - 1 ) + "." + ext;
      QString destname =( trunk+"-backup-%1" ).arg( dest_i ) + "." + ext;
      my_dir.rename( srcname, destname );
    }

    QString srcname = trunk + "." + ext;
    QString destname =( trunk+"-backup-%1" ).arg( 1 ) + "." + ext;
    my_dir.rename( srcname, destname );
  }

  auto_ptr<tProgressDialog> progress( new tProgressDialog( this, false ) );
  progress->setWhat( tr( "Saving database..." ) );
  Database->save( name, progress.get() );
}




void tMainWindow::fileSave()
{
  fileSaveWithResult();
}





bool tMainWindow::fileSaveWithResult()
{
  if ( !FilenameValid )
  {
    return fileSaveAsWithResult();
  }

  try
  {
    saveDatabase( CurrentFilename );
    return true;
  }
  catch ( exception &ex )
  {
    QMessageBox::warning( this, tr( "madman" ),
	tr( "Error saving file '%1'" ).arg( string2QString( ex.what() ) ),
	QMessageBox::Ok, 0 );
    return false;
  }
}




void tMainWindow::fileSaveAs()
{
  fileSaveAsWithResult();
}




bool tMainWindow::fileSaveAsWithResult()
{
  try
  {
    QString filename;
    while ( true )
    {
      filename = QFileDialog::getSaveFileName( QString::null, "madman Database (*.mad)", this);
      if ( filename.isNull() )
	return false;

      QFileInfo info( filename );
      if ( info.exists() )
      {
	int result = QMessageBox::warning( this, tr( "Warning" ), 
	      tr( "The file '%1' exists. Overwrite?" ).arg( filename ), 
	      QMessageBox::Yes, QMessageBox::No | QMessageBox::Default,
	      QMessageBox::Cancel );
	if ( result == QMessageBox::Yes )
	  break;
	if ( result == QMessageBox::Cancel )
	  return false;
      }
      else
	break;
    }
    saveDatabase( filename );
    CurrentFilename = filename;
    FilenameValid = true;
    updateWindowCaption();
    return true;
  }
  catch ( exception &ex )
  {
    QMessageBox::warning( this, tr( "madman" ),
	tr( "Error saving file '%1'" ).arg( string2QString( ex.what() ) ),
	QMessageBox::Ok, 0 );
    return false;
  }
}





void tMainWindow::filePreferences()
{
  if ( editPreferences( this, Preferences, Database->DirectoryList, Settings ) )
  {
    setupListView( AllSongsView, true );
    setupListView( SongSetEditor, true );

    AllSongsView.redisplay( false );
    SongSetEditor.redisplay( false );

    AllSongsView.setup();
    SongSetEditor.setup();

    rescan();
  }
}




void tMainWindow::helpAbout()
{
  QMessageBox::about( this, tr( "About madman" ),
      tr( "<b>madman %1</b><p>"
      "madman administrates digital music archives nicely<br>"
      "Visit the madman homepage at <a href=\"http://madman.sf.net\">http://madman.sf.net</a><p>"
      "Brought to you by<br>"
      "<ul>"
      "  <li>Andreas Kl&ouml;ckner <a href=\"mailto:inducer@users.sf.net\">inducer@users.sf.net</a><p>"
      "</ul>"
      "Some code borrowed from<br>"
      "<ul>"
      "  <li>XMMS &lt;http://www.xmms.org&gt;"
      "  <li>Cedric Tefft &lt;cedric@earthling.net&gt;"
      "  <li>Ricard Cerqueira &lt;rmc@rccn.net&gt;"
      "  <li>Johannes Overmann &lt;overmann@iname.com&gt;"
      "</ul>"
      "(C)opyright 2002,2003. Proudly distributed under the GNU "
      "General Public License." ).arg( MADMAN_VERSION )
      );
}




void tMainWindow::help()
{
  tHelpBrowser *hb = new tHelpBrowser( this );
  hb->browserHelp->setTextFormat( Qt::RichText );
  hb->browserHelp->setText(
      tr(
	"<h2>Searching in madman</h2>"
	"Criteria can be combined via parentheses, <tt>|</tt> (or), <tt>&</tt> (and), "
	"<tt>!</tt> (not). "
	"Plain text will be interpreted as <tt>~any(...)</tt>. "
	"Unconnected criteria will automatically be connected by \"and\". "
	"Comments can be included if enclosed in <tt>{ ... }</tt>. "
	"<h3>Text criteria</h3>"
	"The following criteria exist:<p>"
	"<ul>"
	"<li><tt>~all</tt>: selects all songs."
	"<li><tt>~none</tt>: selects no songs."
	"<li><tt>~any(...)</tt>: selects all songs that contain ... in any field."
	"<li><tt>~artist(...)</tt>: selects all songs that contain ... in the artist field."
	"<li><tt>~performer(...)</tt>: selects all songs that contain ... in the performer field."
	"<li><tt>~album(...)</tt>: selects all songs that contain ... in the album field."
	"<li><tt>~title(...)</tt>: selects all songs that contain ... in the title field."
	"<li><tt>~genre(...)</tt>: selects all songs that contain ... in the genre field."
	"<li><tt>~filename(...)</tt>: selects all songs that contain ... in the filename field."
	"</ul>"
	"Texts mentioned above as '...' can take several match modifiers:<p>"
	"<ul>"
	"<li><tt>re:...</tt> for a regular expression"
	"<li><tt>complete:...</tt> for a complete match"
	"<li><tt>substring:...</tt> for a substring match"
	"<li><tt>fuzzy:...</tt> for a fuzzy match"
	"</ul>"
	"Substring matching is assumed if nothing is specified.<p>"

	"<h3>Numeric criteria</h3>"
	"The following criteria exist:<p>"
	"<ul>"
	"<li><tt>~rating(&lt;|&gt;|&lt;=|=|&gt;=...)</tt>: selects all songs that have a a rating less/greater/.. than ..."
	"<li><tt>~playcount(&lt;|&gt;|&lt;=|=|&gt;=...)</tt>: selects all songs that have a a play count less/greater/.. than ..."
	"<li><tt>~existedfordays(&lt;|&gt;|&lt;=|=|&gt;=...)</tt>: selects all songs that have been known to madman "
	"for less/greater/.. than ... days"
	"<li><tt>~fullplayratio(&lt;|&gt;|&lt;=|=|&gt;=...)</tt>: selects all songs that have a ratio "
	"of being played in full to the total play count less/greater/.. than ..."
	"</ul>"
	"Equal matching is assumed if nothing is specified.<p>"
       ) );
  hb->show();
}




void tMainWindow::addSongSet( tSongSetNode *node )
{
  if ( Database->SongSetTree == NULL )
    Database->SongSetTree = node;
  else if ( currentNode() == NULL )
    Database->SongSetTree->addChild( node );
  else
    currentNode()->addChild( node );

  updateSongSetTree( true, true );
}




void tMainWindow::addSongSet()
{
  tSongSetNode *node = new tSongSetNode( new tSongSet() );
  node->data()->setSongCollection( &Database->SongCollection );
  node->setName( tr( "New Playlist" ) );
  addSongSet( node );
}




void tMainWindow::duplicateSongSet()
{
  tSongSetNode *current_node = currentNode();
  if ( current_node == NULL )
    return;

  tSongSetNode *new_node = new tSongSetNode( new tSongSet );
  new_node->setName( tr( "Copy of %1" ).arg( current_node->name() ) );
  *(new_node->data()) = *(current_node->data());
  current_node->addChild( new_node );

  updateSongSetTree( true, true );
}




void tMainWindow::removeSongSet()
{
  if ( !lstSets->currentItem( ) )
    return ;

  tSongSetNode *node = currentNode();
  if ( node->begin() != node->end() )
  {
    if ( QMessageBox::warning( this, tr( "Warning" ), 
	  tr( "This playlist has children. Do you still want to delete it?" ), 
	  QMessageBox::Yes, QMessageBox::No | QMessageBox::Default ) == QMessageBox::No )
      return ;
  }

  // delete from structure
  if ( node->parent() == NULL )
    Database->SongSetTree = NULL;
  else
    node->parent()->removeChild( node );
  delete node;

  SongSetEditor.setSongSet( NULL );

  updateSongSetTree( false, true );
}




void tMainWindow::editMultilineSongSetCriterion()
{
  if ( !lstSets->currentItem( ) )
    return ;
  tSongSetNode *node = currentNode();
  QString criterion = node->data()->criterion();
  while ( multilineEdit( tr( "madman" ), 
	tr( "Criterion for dynamic playlist" ), criterion, this ) )
  {
    try
    {
      if ( node->data()->criterion() != criterion )
      {
	node->data()->setCriterion( criterion );
	node->data()->reevaluateCriterion();
	songSetSelectionChanged();
      }
      break;
    }
    catch ( exception &ex )
    {
      QMessageBox::warning( this, tr( "madman" ),
	tr( "Error in Criterion: %1" ).arg( string2QString( ex.what() ) ),
	QMessageBox::Ok, 0 );
    }
  }
}




void tMainWindow::slotDropSongSetNode( const QString &node_name, tSongSetNode *onto )
{

  tSongSetNode *node = Database->SongSetTree->resolve( node_name );
  if ( node == NULL )
    return;

  // this also catches the case when root is being dragged away
  if ( node == onto || onto->hasParent( node ) )
    return;

  // by now, node can't be root, so it has a parent.
  node->parent()->removeChild( node );
  onto->insertChild( node, onto->begin() );

  updateSongSetTree( false, true );
}




void tMainWindow::renameSongSet( QListViewItem *item, int col, const QString &text )
{
  try
  {
    nodeFromItem( item )->setName( text );
  }
  catch ( exception &ex )
  {
    QMessageBox::warning( this, tr( "madman" ), string2QString( ex.what() ), QMessageBox::Ok, 0 );
  }
  updateSongSetTree( true, true );
}




void tMainWindow::songSetSelectionChanged()
{
  tSongSetNode *node = currentNode();

  if ( node )
  {
    tSongSet *song_set = node->data();
    SongSetEditor.setSongSet( song_set );
    editSongSetCriterion->setText( song_set->criterion() );
    editSongSetCriterion->setEnabled( true );
  }
  else
  {
    editSongSetCriterion->setEnabled( false );
    SongSetEditor.setSongSet( NULL );
  }
}




void tMainWindow::songSetCriterionChanged()
{
  tSongSetNode *node = currentNode();

  if ( node )
  {
    tSongSet *song_set = node->data();
    try {
      QString newcrit = editSongSetCriterion->text();
      if ( song_set->criterion() != newcrit )
      {
	song_set->setCriterion( newcrit );
	song_set->reevaluateCriterion();
      }
    }
    catch ( exception &ex )
    {
      QMessageBox::warning( this, tr( "madman" ),
	tr( "Error in Criterion: %1" ).arg( string2QString( ex.what() ) ),
	QMessageBox::Ok, 0 );
    }
  }
}




// set contents management ----------------------------------------------------
void tMainWindow::add( const tSongList &songlist )
{
  FOREACH_CONST( first, songlist, tSongList )
    SongSetEditor.songSet()->add( *first );
}




void tMainWindow::searchChanged()
{
  try 
  {
    QString new_crit = editSearch->text();
    if ( AllSongsSet.criterion() != new_crit )
    {
      if ( CurrentSearchProgress )
	CurrentSearchProgress->cancel();

      bool prev_crit_was_binary = AllSongsSet.hasBinaryCriterion();

      AllSongsSet.setCriterion( new_crit, Preferences.UseFuzzySearch );

      if ( !AllSongsSet.hasBinaryCriterion() )
      {
	AllSongsView.setDisplaySorting( FIELD_EXTRA_CRITERION_MATCH, false );
      }
      else
      {
	if ( !prev_crit_was_binary )
	  AllSongsView.setDisplaySorting( FIELD_ARTIST, true );
      }

      tCancellableStatusBarProgress progress( statusBar() );
      CurrentSearchProgress = &progress;

      AllSongsSet.reevaluateCriterion( &progress );

      if ( CurrentSearchProgress == &progress )
	CurrentSearchProgress = NULL;
    }
  }
  catch ( exception &ex )
  {
    statusBar()->message( tr( "Error in expression: %1." ).arg( string2QString( ex.what() ) ), 2000 );
  }
}




void tMainWindow::buildOverviewTree()
{
  lstOverview->clear();

  new tGenreOverviewItem( *Database, lstOverview, NULL, tr( "By Genre" ), QString::null );
  new tAlbumOverviewItem( *Database, lstOverview, NULL, tr( "By Album" ), QString::null );
  new tArtistsOverviewItem( *Database, lstOverview, NULL, tr( "By Artist" ), QString::null );

  new tOverviewItem( lstOverview, NULL, tr( "All" ), "" );
}




void tMainWindow::overviewSelectionChanged( QListViewItem *item )
{
  tOverviewItem *ov_item = dynamic_cast< tOverviewItem * >( item );
  if ( ov_item->criterion() != QString::null )
    editSearch->setText( ov_item->criterion() );
}




void tMainWindow::checkCurrentlyPlayingSong()
{
  if ( Preferences.Player && Preferences.CollectHistory )
  {
    QString song_file = Preferences.Player->currentFilename();

    if ( Preferences.Player->isPlaying() )
      CurrentSongPlayingSeconds += 0.5;

    if ( song_file != CurrentSongFilename )
    {
      if ( CurrentSongFilename != QString::null )
      {
	tSong *old_song = Database->SongCollection.getByFilename( CurrentSongFilename );
	if ( old_song )
	{
	  old_song->played( time( NULL ), CurrentSongPlayingSeconds > 0.6 * old_song->duration() );
	  Database->History.played( old_song->uniqueId(), 
	      time( NULL ), CurrentSongPlayingSeconds );
	}
      }
      
      cout << "song changed: " << song_file << endl;

      CurrentSongFilename = song_file;
      CurrentSongPlayingSeconds = 0;
    }
  }
}




// helpers --------------------------------------------------------------------
void tMainWindow::updateSongSetTree( bool keep_selection, bool keep_scroll )
{
  int contents_x = lstSets->contentsX();
  int contents_y = lstSets->contentsY();
  tSongSetNodeListViewItem *selected_item = NULL;
  if ( keep_selection )
    selected_item = dynamic_cast<tSongSetNodeListViewItem *>( lstSets->currentItem() );
  tSongSetNode *selected_node = NULL;
  if ( selected_item )
    selected_node = selected_item->node();

  lstSets->clear();
  if ( Database->SongSetTree )
    lstSets->insertItem( 
	createSongSetNodeItem( Database->SongSetTree, NULL, 
	  selected_node, selected_item ) );

  if ( keep_scroll )
    lstSets->setContentsPos( contents_x, contents_y );
  if ( keep_selection && selected_item )
    lstSets->setSelected( selected_item, true );
}




tSongSetNode *tMainWindow::nodeFromItem( QListViewItem *item )
{
  tSongSetNodeListViewItem *cast_item = dynamic_cast<tSongSetNodeListViewItem *>( item );
  if ( cast_item )
    return cast_item->node();
  else
    return NULL;
}





tSongSetNode *tMainWindow::currentNode()
{
  QListViewItem *current_item = lstSets->currentItem();
  if ( current_item == NULL )
    return NULL;
  else
    return nodeFromItem( current_item );
}




tSongSet *tMainWindow::currentSongSet()
{
  tSongSetNode * node = currentNode();
  if ( node )
    return node->data();
  else
    return NULL;
}




void tMainWindow::insertAtPosition( QListViewItem *parent, QListViewItem *item, tIndex pos )
{
  parent->insertItem( item );
  if ( pos != 0 )
  {
    QListViewItem * after = item;
    while ( pos-- )
      after = after->nextSibling();
    item->moveItem( after );
  }
}




QListViewItem *tMainWindow::createSongSetNodeItem( tSongSetNode *node, QListViewItem *parent,
    tSongSetNode *selected_node, tSongSetNodeListViewItem *&selected_item )
{
  tSongSetNodeListViewItem *result;
  if ( parent )
    result = new tSongSetNodeListViewItem( parent, node );
  else
    result = new tSongSetNodeListViewItem( lstSets, node );

  connect( result, SIGNAL(dropNode( const QString &,tSongSetNode *)),
      this, SLOT( slotDropSongSetNode( const QString &,tSongSetNode *) ) );

  if ( node == selected_node )
    selected_item = result;

  tSongSetNode::iterator first = node->begin(), last = node->end();
  if ( first != last )
    do
    {
      last--;
      createSongSetNodeItem( *last, result, selected_node, selected_item );
    }
    while ( first != last );

  lstSets->setOpen( result, true );
  return result;
}




void tMainWindow::loadSplitterAppearance( QSplitter *splitter, QSettings &settings, QString const &where )
{
  // rewriting this code for Qt 3.0.x is not worth it.

#if !( QT_VERSION >> 16 == 3 && ((QT_VERSION >> 8) & 0xff) == 0 )
  QString representation;
  representation = settings.readEntry( where );
  {
    QTextStream stream( &representation, IO_ReadOnly );
    stream >> *splitter;
  }
#endif
}




void tMainWindow::saveSplitterAppearance( QSplitter *splitter, QSettings &settings, QString const &where )
{
  // rewriting this code for Qt 3.0.x is not worth it.

#if !( QT_VERSION >> 16 == 3 && ((QT_VERSION >> 8) & 0xff) == 0 )
  QString representation;
  {
    QTextStream stream( &representation, IO_WriteOnly );
    stream << *splitter;
  }
  settings.writeEntry( where, representation );
#endif
}

