#include <qlistbox.h>
#include <qpushbutton.h>
#include <qcombobox.h>
#include <qcheckbox.h>
#include <qtabwidget.h>
#include <qspinbox.h>
#include <qlistview.h>
#include <qlabel.h>

#include <stdexcept>

#include "utility/player.h"
#include "utility/plugin.h"
#include "ui/mainwin.h"

#include "prefs.moc"




// tPreferences ---------------------------------------------------------------
void tPreferences::save( QSettings &settings )
{
  if ( Player )
  {
    settings.writeEntry( "/madman/player", Player->name() );
  }

  QStringList plugin_dirs = QStringvector2QStringList( PluginDirectories );
  settings.writeEntry( "/madman/plugin_directories", plugin_dirs );
  settings.writeEntry( "/madman/remember_geometry", RememberGeometry ? 1 : 0 );
  settings.writeEntry( "/madman/scan_at_startup", ScanAtStartup ? 1 : 0 );
  settings.writeEntry( "/madman/collect_history", CollectHistory ? 1 : 0 );
  settings.writeEntry( "/madman/use_fuzzy_search", UseFuzzySearch ? 1 : 0 );
  settings.writeEntry( "/madman/number_of_backups_kept", BackupCount );

  for ( int i = 0; i < FIELD_COUNT; i++ )
    settings.writeEntry( 
	QString( "/madman/column.%1.visible" ).arg( i ), 
	ColumnVisibility[i] ? 1 : 0 );
}




void tPreferences::load( QSettings &settings )
{
  QString player_name = settings.readEntry( "/madman/player" );
  
  tPlayer * newplayer = createPlayer( player_name );
  setPlayer( newplayer );
  if ( !Player )
  {
    vector<QString> players;
    listPlayers( players );
    if ( players.size() )
      setPlayer( createPlayer( players[0] ) );
    if ( !Player )
      throw runtime_error( "can't instantiate any player, giving up." );
  }

  bool ok;
  PluginDirectories = QStringList2QStringvector( settings.readListEntry( "/madman/plugin_directories", &ok ) );
  if ( !ok )
  {
    PluginDirectories.clear();
    PluginDirectories.push_back( MADMAN_LIBDIR "/madman/plugins" );
  }
  RememberGeometry = settings.readNumEntry( "/madman/remember_geometry", 1 );
  ScanAtStartup = settings.readNumEntry( "/madman/scan_at_startup", 0 );
  CollectHistory = settings.readNumEntry( "/madman/collect_history", 1 );
  UseFuzzySearch = settings.readNumEntry( "/madman/use_fuzzy_search", 1 );
  BackupCount = settings.readNumEntry( "/madman/number_of_backups_kept", 4 );

  ColumnVisibility.resize( FIELD_COUNT );
  for ( int i = 0; i < FIELD_COUNT; i++ )
  {
    int default_vis = 1;
    if ( i == FIELD_FULLPLAYCOUNT ) default_vis = 0;
    if ( i == FIELD_PARTIALPLAYCOUNT ) default_vis = 0;
    if ( i == FIELD_PERFORMER ) default_vis = 0;
    if ( i == FIELD_EXTRA_CRITERION_MATCH ) default_vis = 0;
    if ( i == FIELD_FILE ) default_vis = 0;
    if ( i == FIELD_PATH ) default_vis = 0;
    if ( i == FIELD_SIZE ) default_vis = 0;

    ColumnVisibility[i] = settings.readNumEntry( 
	QString( "/madman/column.%1.visible" ).arg( i ), 
	default_vis ) != 0;
  }
}




void tPreferences::setPlayer( tPlayer *newplayer )
{
  if ( newplayer )
  {
    if ( Player ) 
      delete Player;
    Player = newplayer;
  }
}




// tPreferencesDialog ---------------------------------------------------------



tPreferencesDialog::tPreferencesDialog( QWidget* parent, tDirectoryList &media_dir_list, 
    tDirectoryList &plugin_dir_list, QSettings &settings )
: tPreferencesDialogBase( parent, "prefs", true ),
  MediaDirectoryListManager( this, media_dir_list, lstMediaDirectories, btnAddMediaDirectory, btnRemoveMediaDirectory ),
  PluginDirectoryListManager( this, plugin_dir_list, lstPluginDirectories, btnAddPluginDirectory, btnRemovePluginDirectory ),
  Settings( settings )
{
  connect( &PluginDirectoryListManager, SIGNAL( changed() ),
      this, SLOT( updatePluginList() ) );
  connect( lstPlugins, SIGNAL( highlighted( int ) ),
      this, SLOT( updateSelectedPluginInfo( int ) ) );
  connect( btnConfigurePlugin, SIGNAL( clicked() ),
      this, SLOT( configurePlugin() ) );
  updatePluginList();
}




void tPreferencesDialog::updatePluginList()
{
  tPluginList plugins;
  enumeratePlugins( PluginDirectoryListManager.directoryList(),
      plugins );

  lstPlugins->clear();
  FOREACH_CONST( first, plugins, tPluginList )
    lstPlugins->insertItem( first->Name );
}




void tPreferencesDialog::updateSelectedPluginInfo( int index )
{
  tPluginList plugins;
  enumeratePlugins( PluginDirectoryListManager.directoryList(),
      plugins );

  if ( index < (int) plugins.size() )
  {
    SelectedPlugin = plugins[ index ].Filename;
    lblPluginName->setText( plugins[ index ].Name );
    lblPluginDescription->setText( plugins[ index ].Description );
  }
}




void tPreferencesDialog::configurePlugin()
{
  tPluginList plugins;
  enumeratePlugins( PluginDirectoryListManager.directoryList(),
      plugins );
  FOREACH( first, plugins, tPluginList )
    if ( first->Filename == SelectedPlugin )
    {
      first->configure( Settings );
      return;
    }
}




// public functions -----------------------------------------------------------
bool editPreferences( tMainWindow *mainwin, tPreferences &prefs, 
    tDirectoryList &media_dir_list, 
    QSettings &settings, int starting_tab )
{
  // create the dialog
  tDirectoryList work_media_dir_list,work_plugin_dir_list;
  work_media_dir_list = media_dir_list;
  work_plugin_dir_list = prefs.PluginDirectories;

  tPreferencesDialog prefsdialog( mainwin, 
      work_media_dir_list, 
      work_plugin_dir_list, settings );
  prefsdialog.tabMain->setCurrentPage( starting_tab );

  // populate the dialog
  vector<QString> players;
  listPlayers( players );
  int index = 0,current_index = 0;
  FOREACH_CONST( first, players, vector<QString> )
  {
    if ( prefs.Player && prefs.Player->name() == *first )
      current_index = index;
    prefsdialog.comboPlayer->insertItem( *first );
    index++;
  }
  prefsdialog.comboPlayer->setCurrentItem( current_index );

  prefsdialog.checkScanAtStartup->setChecked(  prefs.ScanAtStartup );
  prefsdialog.checkRememberGeometry->setChecked( prefs.RememberGeometry );
  prefsdialog.checkHistory->setChecked( prefs.CollectHistory );
  prefsdialog.checkUseFuzzy->setChecked( prefs.UseFuzzySearch );
  prefsdialog.spinBackupCount->setValue( prefs.BackupCount );

  prefsdialog.lstVisibleColumns->setSorting( -1 );
  QCheckListItem *last = NULL;
  for ( int i = 0; i < FIELD_COUNT; i++ )
  {
    last = new QCheckListItem( prefsdialog.lstVisibleColumns, last,
	getFieldName( (tSongField) i ), QCheckListItem::CheckBox );
    last->setOn( prefs.ColumnVisibility[ i ] );
  }

  // show
  int returncode = prefsdialog.exec();

  // process results
  if ( returncode == QDialog::Accepted )
  {
    media_dir_list = work_media_dir_list;
    prefs.PluginDirectories = work_plugin_dir_list;

    prefs.setPlayer( createPlayer( prefsdialog.comboPlayer->currentText() ) );

    prefs.ScanAtStartup = prefsdialog.checkScanAtStartup->isChecked();
    prefs.RememberGeometry = prefsdialog.checkRememberGeometry->isChecked();
    prefs.CollectHistory = prefsdialog.checkHistory->isChecked();
    prefs.UseFuzzySearch = prefsdialog.checkUseFuzzy->isChecked();
    prefs.BackupCount = prefsdialog.spinBackupCount->value();

    QListViewItem *item = prefsdialog.lstVisibleColumns->firstChild();
    for ( int i = 0; i < FIELD_COUNT; i++ )
    {
      prefs.ColumnVisibility[ i ] = dynamic_cast<QCheckListItem *>(item)->isOn();
      item = item->nextSibling();
    }
    return true;
  }
  else
    return false;
}
